@@ -9,7 +9,7 @@
 # see http://specs.xmlsoap.org/ws/2005/04/discovery/ws-discovery.pdf and
 # related documents for details (look at README for more references)
 #
-# (c) Steffen Christgau, 2017-2020
+# (c) Steffen Christgau, 2017-2021
 
 import sys
 import signal
@@ -35,6 +35,8 @@ import pwd
 import grp
 import datetime
 
+WSDD_VERSION = '0.6.4'
+
 
 class MulticastHandler:
     """
@@ -43,16 +45,30 @@ class MulticastHandler:
     """
     # TODO: this one needs some cleanup
     def __init__(self, family, address, interface, selector):
+        # network address, family, and interface name
         self.address = address
         self.family = family
         self.interface = interface
+
+        # create individual interface-bound sockets for:
+        #  - receiving multicast traffic
+        #  - sending multicast from a socket bound to WSD port
+        #  - sending unicast messages from a random port
         self.recv_socket = socket.socket(self.family, socket.SOCK_DGRAM)
         self.recv_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
-        self.send_socket = socket.socket(self.family, socket.SOCK_DGRAM)
+        self.mc_send_socket = socket.socket(self.family, socket.SOCK_DGRAM)
+        self.uc_send_socket = socket.socket(self.family, socket.SOCK_DGRAM)
+
+        # the string representation of the local address
+        # overridden in network setup (for IPv6)
         self.transport_address = address
+
+        # family-specific multicast address and local HTTP server (tuples)
         self.multicast_address = None
         self.listen_address = None
 
+        # dictionary that holds objects with a handle_request method for the
+        # sockets created above
         self.message_handlers = {}
         self.selector = selector
 
@@ -68,15 +84,19 @@ class MulticastHandler:
         logger.debug('will listen for HTTP traffic on address {0}'.format(
             self.listen_address))
 
+        # register calbacks for incoming data (also for mc)
         self.selector.register(self.recv_socket, selectors.EVENT_READ, self)
-        self.selector.register(self.send_socket, selectors.EVENT_READ, self)
+        self.selector.register(self.mc_send_socket, selectors.EVENT_READ, self)
+        self.selector.register(self.uc_send_socket, selectors.EVENT_READ, self)
 
     def cleanup(self):
         self.selector.unregister(self.recv_socket)
-        self.selector.unregister(self.send_socket)
+        self.selector.unregister(self.mc_send_socket)
+        self.selector.unregister(self.uc_send_socket)
 
         self.recv_socket.close()
-        self.send_socket.close()
+        self.uc_send_socket.close()
+        self.mc_send_socket.close()
 
     def handles(self, family, addr, interface):
         return (self.family == family and self.address == addr and
@@ -95,6 +115,16 @@ class MulticastHandler:
         self.recv_socket.setsockopt(
             socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
 
+        # Could anyone ask the Linux folks for the rationale for this!?
+        if platform.system() == 'Linux':
+            try:
+                # supported starting from Linux 4.20
+                IPV6_MULTICAST_ALL = 29
+                self.recv_socket.setsockopt(
+                    socket.IPPROTO_IPV6, IPV6_MULTICAST_ALL, 0)
+            except OSError as e:
+                logger.warning('cannot unset all_multicast: {}'.format(e))
+
         # bind to network interface, i.e. scope and handle OS differences,
         # see Stevens: Unix Network Programming, Section 21.6, last paragraph
         try:
@@ -102,11 +132,14 @@ class MulticastHandler:
         except OSError:
             self.recv_socket.bind(('::', 0, 0, idx))
 
-        self.send_socket.setsockopt(
+        # bind unicast socket to interface address and ephemeral port
+        self.uc_send_socket.bind((self.address, WSD_UDP_PORT, 0, idx))
+
+        self.mc_send_socket.setsockopt(
             socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_LOOP, 0)
-        self.send_socket.setsockopt(
+        self.mc_send_socket.setsockopt(
             socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_HOPS, args.hoplimit)
-        self.send_socket.setsockopt(
+        self.mc_send_socket.setsockopt(
             socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_IF, idx)
 
         self.transport_address = '[{0}]'.format(self.address)
@@ -124,16 +157,23 @@ class MulticastHandler:
         self.recv_socket.setsockopt(
             socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
 
+        if platform.system() == 'Linux':
+            IP_MULTICAST_ALL = 49
+            self.recv_socket.setsockopt(socket.IPPROTO_IP, IP_MULTICAST_ALL, 0)
+
         try:
             self.recv_socket.bind((WSD_MCAST_GRP_V4, WSD_UDP_PORT))
         except OSError:
             self.recv_socket.bind(('', WSD_UDP_PORT))
 
-        self.send_socket.setsockopt(
+        # bind unicast socket to interface address and WSD's udp port
+        self.uc_send_socket.bind((self.address, WSD_UDP_PORT))
+
+        self.mc_send_socket.setsockopt(
             socket.IPPROTO_IP, socket.IP_MULTICAST_IF, mreq)
-        self.send_socket.setsockopt(
+        self.mc_send_socket.setsockopt(
             socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, 0)
-        self.send_socket.setsockopt(
+        self.mc_send_socket.setsockopt(
             socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, args.hoplimit)
 
         self.listen_address = (self.address, WSD_HTTP_PORT)
@@ -157,18 +197,29 @@ class MulticastHandler:
 
     def handle_request(self, key):
         s = None
-        if key.fileobj == self.send_socket:
-            s = self.send_socket
+        if key.fileobj == self.uc_send_socket:
+            s = self.uc_send_socket
+        elif key.fileobj == self.mc_send_socket:
+            s = self.mc_send_socket
         elif key.fileobj == self.recv_socket:
             s = self.recv_socket
         else:
-            return
+            raise ValueError("Unknown socket passed as key.")
 
         msg, address = s.recvfrom(WSD_MAX_LEN)
         if s in self.message_handlers:
             for handler in self.message_handlers[s]:
                 handler.handle_request(msg, address)
 
+    def send(self, msg, addr):
+        # Request from a client must be answered from a socket that is bound
+        # to the WSD port, i.e. the recv_socket. Messages to multicast
+        # addresses are sent over the dedicated send socket.
+        if addr == self.multicast_address:
+            self.mc_send_socket.sendto(msg, addr)
+        else:
+            self.uc_send_socket.sendto(msg, addr)
+
 
 # constants for WSD XML/SOAP parsing
 WSA_URI = 'http://schemas.xmlsoap.org/ws/2004/08/addressing'
@@ -320,16 +371,22 @@ class WSDMessageHandler(object):
         """
         tree = ElementTree.fromstring(msg)
         header = tree.find('./soap:Header', namespaces)
-        msg_id = header.find('./wsa:MessageID', namespaces).text
+        msg_id_tag = header.find('./wsa:MessageID', namespaces)
+        if msg_id_tag is None:
+            return None
+
+        msg_id = msg_id_tag.text
 
         # if message came over a MulticastHandler, check for duplicates
         if mch and self.is_duplicated_msg(msg_id):
             logger.debug('known message ({0}): dropping it'.format(msg_id))
             return None
 
-        response = None
-        action = header.find('./wsa:Action', namespaces).text
-        body = tree.find('./soap:Body', namespaces)
+        action_tag = header.find('./wsa:Action', namespaces)
+        if action_tag is None:
+            return None
+
+        action = action_tag.text
         _, _, action_method = action.rpartition('/')
 
         if mch:
@@ -342,6 +399,10 @@ class WSDMessageHandler(object):
             logger.debug('processing WSD {} message ({})'.format(
                 action_method, msg_id))
 
+        body = tree.find('./soap:Body', namespaces)
+        if body is None:
+            return None
+
         logger.debug('incoming message content is {0}'.format(msg))
         if action in self.handlers:
             handler = self.handlers[action]
@@ -420,6 +481,7 @@ class WSDDiscoveredDevice(object):
         self.last_seen = None
         self.addresses = {}
         self.props = {}
+        self.display_name = ''
 
         self.update(xml_str, xaddr, interface)
 
@@ -430,15 +492,15 @@ class WSDDiscoveredDevice(object):
         for section in sections:
             dialect = section.attrib['Dialect']
             if dialect == WSDP_URI + '/ThisDevice':
-                self.extract_wsdp_props(section, self.props, dialect)
+                self.extract_wsdp_props(section, dialect)
             elif dialect == WSDP_URI + '/ThisModel':
-                self.extract_wsdp_props(section, self.props, dialect)
+                self.extract_wsdp_props(section, dialect)
             elif dialect == WSDP_URI + '/Relationship':
                 host_xpath = ('wsdp:Relationship[@Type="{}/host"]/wsdp:Host'
                               .format(WSDP_URI))
                 host_sec = section.find(host_xpath, namespaces)
                 if (host_sec):
-                    self.extract_host_props(host_sec, self.props)
+                    self.extract_host_props(host_sec)
             else:
                 logger.debug('unknown metadata dialect ({})'.format(dialect))
 
@@ -450,12 +512,19 @@ class WSDDiscoveredDevice(object):
             self.addresses[interface].add(addr)
 
         self.last_seen = time.time()
-        logger.info('discovered {} in {} on {}%{}'.format(
-            self.props['DisplayName'], self.props['BelongsTo'], addr,
-            interface.interface.name))
+        if ('DisplayName' in self.props) and ('BelongsTo' in self.props):
+            self.display_name = self.props['DisplayName']
+            logger.info('discovered {} in {} on {}%{}'.format(
+                self.display_name, self.props['BelongsTo'], addr,
+                interface.interface.name))
+        elif 'FriendlyName' in self.props:
+            self.display_name = self.props['FriendlyName']
+            logger.info('discovered {} on {}%{}'.format(
+                self.display_name, addr, interface.interface.name))
+
         logger.debug(str(self.props))
 
-    def extract_wsdp_props(self, root, target_dict, dialect):
+    def extract_wsdp_props(self, root, dialect):
         _, _, propsRoot = dialect.rpartition('/')
         # XPath support is limited, so filter by namespace on our own
         nodes = root.findall('./wsdp:{0}/*'.format(propsRoot), namespaces)
@@ -463,16 +532,16 @@ class WSDDiscoveredDevice(object):
         prop_nodes = [n for n in nodes if n.tag.startswith(ns_prefix)]
         for node in prop_nodes:
             tag_name = node.tag[len(ns_prefix):]
-            target_dict[tag_name] = node.text
+            self.props[tag_name] = node.text
 
-    def extract_host_props(self, root, target_dict):
+    def extract_host_props(self, root):
         types = root.findtext('wsdp:Types', '', namespaces)
-        target_dict['types'] = types.split(' ')
+        self.props['types'] = types.split(' ')
         if types != PUB_COMPUTER:
             return
 
         comp = root.findtext(PUB_COMPUTER, '', namespaces)
-        target_dict['DisplayName'], _, target_dict['BelongsTo'] = (
+        self.props['DisplayName'], _, self.props['BelongsTo'] = (
             comp.partition('/'))
 
 
@@ -481,7 +550,7 @@ class WSDClient(WSDUDPMessageHandler):
     def __init__(self, mch, known_devices):
         super().__init__(mch)
 
-        self.mch.add_handler(self.mch.send_socket, self)
+        self.mch.add_handler(self.mch.mc_send_socket, self)
         self.mch.add_handler(self.mch.recv_socket, self)
 
         self.probes = {}
@@ -497,7 +566,7 @@ class WSDClient(WSDUDPMessageHandler):
         self.send_probe()
 
     def cleanup(self):
-        self.mch.remove_handler(self.mch.send_socket, self)
+        self.mch.remove_handler(self.mch.mc_send_socket, self)
         self.mch.remove_handler(self.mch.recv_socket, self)
 
     def send_probe(self):
@@ -567,7 +636,7 @@ class WSDClient(WSDUDPMessageHandler):
         rm_path = 'wsd:ResolveMatches/wsd:ResolveMatch'
         endpoint, xaddrs = self.extract_endpoint_metadata(body, rm_path)
         if not endpoint or not xaddrs:
-            logger.warning('resolve match without endpoint/xaddr')
+            logger.debug('resolve match without endpoint/xaddr')
             return
 
         xaddr = xaddrs.strip()
@@ -605,7 +674,8 @@ class WSDClient(WSDUDPMessageHandler):
             with urllib.request.urlopen(request, None, 2.0) as stream:
                 self.handle_metadata(stream.read(), endpoint, xaddr)
         except urllib.error.URLError as e:
-            logger.warn('could not fetch metadata from: {}'.format(url, e))
+            logger.warning('could not fetch metadata from: {} {}'.format(
+                url, e))
 
     def build_getmetadata_message(self, endpoint):
         tree, _ = self.build_message_tree(endpoint, WSD_GET, None, None)
@@ -684,7 +754,7 @@ class WSDHost(WSDUDPMessageHandler):
 
         if scopes:
             # THINK: send fault message (see p. 21 in WSD)
-            logger.warning('scopes ({}) unsupported but probed'.format(scopes))
+            logger.debug('scopes ({}) unsupported but probed'.format(scopes))
             return None
 
         types_elem = probe.find('./wsd:Types', namespaces)
@@ -713,7 +783,7 @@ class WSDHost(WSDUDPMessageHandler):
             return None
 
         if not addr.text == args.uuid.urn:
-            logger.warning(
+            logger.debug(
                 'invalid resolve request: address ({}) does not '
                 'match own one ({})'.format(addr.text, args.uuid.urn))
             return None
@@ -788,9 +858,11 @@ class WSDHttpServer(http.server.HTTPServer):
     """ HTTP server both with IPv6 support and WSD handling """
 
     def __init__(self, mch, RequestHandlerClass, addr_family, sel):
-        if addr_family == socket.AF_INET6:
-            type(self).address_family = addr_family
+        # hacky way to convince HTTP/SocketServer of the address family
+        type(self).address_family = addr_family
 
+        # remember actual address family used by the server instance
+        self.addr_family = addr_family
         self.mch = mch
         self.selector = sel
         self.wsd_handler = WSDHttpMessageHandler()
@@ -799,10 +871,13 @@ class WSDHttpServer(http.server.HTTPServer):
         super().__init__(mch.listen_address, RequestHandlerClass)
 
     def server_bind(self):
-        if type(self).address_family == socket.AF_INET6:
+        if self.address_family == socket.AF_INET6:
             self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
 
         super().server_bind()
+
+    def server_activate(self):
+        super().server_activate()
         self.selector.register(self.fileno(), selectors.EVENT_READ, self)
         self.registered = True
 
@@ -820,11 +895,12 @@ class WSDHttpRequestHandler(http.server.BaseHTTPRequestHandler):
 
     def do_POST(self):
         if self.path != '/' + str(args.uuid):
-            self.send_error()
+            self.send_error(http.HTTPStatus.NOT_FOUND)
 
         ct = self.headers['Content-Type']
         if ct is None or not ct.startswith(MIME_TYPE_SOAP_XML):
-            self.send_error(http.HTTPStatus.NOT_FOUND, 'Invalid Content-Type')
+            self.send_error(http.HTTPStatus.BAD_REQUEST,
+                            'Invalid Content-Type')
 
         content_length = int(self.headers['Content-Length'])
         body = self.rfile.read(content_length)
@@ -877,8 +953,8 @@ class ApiRequestHandler(socketserver.StreamRequestHandler):
 
             retval = retval + '{}\t{}\t{}\t{}\t{}\n'.format(
                 dev_uuid,
-                dev.props['DisplayName'],
-                dev.props['BelongsTo'],
+                dev.display_name,
+                dev.props['BelongsTo'] if 'BelongsTo' in dev.props else '',
                 datetime.datetime.fromtimestamp(dev.last_seen).isoformat(
                     'T', 'seconds'),
                 ','.join(addrs_str))
@@ -1018,9 +1094,9 @@ class NetworkAddressMonitor(object,  metaclass=MetaEnumAfterInit):
 
     def handle_deleted_address(self, raw_addr, addr_family, interface):
         addr = socket.inet_ntop(addr_family, raw_addr)
-        logger.warn('deleted address {} on {}'.format(addr, interface.name))
+        logger.info('deleted address {} on {}'.format(addr, interface.name))
 
-        if not self.is_address_handled(addr_family, raw_addr, interface):
+        if not self.is_address_handled(raw_addr, addr_family, interface):
             return
 
         mch = self.get_mch_by_address(addr_family, addr, interface)
@@ -1103,6 +1179,10 @@ RTA_ALIGNTO = 4
 RTA_LEN = 4
 
 
+def align_to(x, n):
+    return ((x + n - 1) // n) * n
+
+
 class NetlinkAddressMonitor(NetworkAddressMonitor):
     """
     Implementation of the AddressMonitor for Netlink sockets, i.e. Linux
@@ -1123,7 +1203,7 @@ class NetlinkAddressMonitor(NetworkAddressMonitor):
 
         self.socket = socket.socket(socket.AF_NETLINK, socket.SOCK_RAW,
                                     socket.NETLINK_ROUTE)
-        self.socket.bind((os.getpid(), rtm_groups))
+        self.socket.bind((0, rtm_groups))
         self.selector.register(self.socket, selectors.EVENT_READ, self)
 
     def enumerate(self):
@@ -1131,8 +1211,7 @@ class NetlinkAddressMonitor(NetworkAddressMonitor):
 
         kernel = (0, 0)
         req = struct.pack('@IHHIIB', NLM_HDR_LEN + 1, self.RTM_GETADDR,
-                          NLM_F_REQUEST | NLM_F_DUMP, 1, os.getpid(),
-                          socket.AF_PACKET)
+                          NLM_F_REQUEST | NLM_F_DUMP, 1, 0, socket.AF_PACKET)
         self.socket.sendto(req, kernel)
 
     def handle_request(self):
@@ -1144,6 +1223,7 @@ class NetlinkAddressMonitor(NetworkAddressMonitor):
         offset = 0
         while offset < len(buf):
             h_len, h_type, _, _, _ = struct.unpack_from('@IHHII', buf, offset)
+            offset += NLM_HDR_LEN
 
             msg_len = h_len - NLM_HDR_LEN
             if msg_len < 0:
@@ -1151,18 +1231,17 @@ class NetlinkAddressMonitor(NetworkAddressMonitor):
 
             if h_type != self.RTM_NEWADDR and h_type != self.RTM_DELADDR:
                 logger.debug('invalid rtm_message type {}'.format(h_type))
-                offset += ((msg_len + 1) // NLM_HDR_ALIGNTO) * NLM_HDR_ALIGNTO
+                offset += align_to(msg_len, NLM_HDR_ALIGNTO)
                 continue
 
-            offset += NLM_HDR_LEN
             # decode ifaddrmsg as in rtnetlink.h
             ifa_family, _, ifa_flags, ifa_scope, ifa_idx = struct.unpack_from(
                 '@BBBBI', buf, offset)
             if (ifa_flags & IFA_F_DADFAILED or ifa_flags & IFA_F_HOMEADDRESS or
                ifa_flags & IFA_F_DEPRECATED or ifa_flags & IFA_F_TENTATIVE):
-                logger.warn('ignore address with invalid state {}'.format(
+                logger.debug('ignore address with invalid state {}'.format(
                     hex(ifa_flags)))
-                offset += ((msg_len + 1) // NLM_HDR_ALIGNTO) * NLM_HDR_ALIGNTO
+                offset += align_to(msg_len, NLM_HDR_ALIGNTO)
                 continue
 
             addr = None
@@ -1184,21 +1263,24 @@ class NetlinkAddressMonitor(NetworkAddressMonitor):
                     addr = buf[i + 4:i + 4 + 16]
                 elif attr_type == IFA_FLAGS:
                     _, ifa_flags = struct.unpack_from('HI', buf, i)
-                i += ((attr_len + 1) // RTA_ALIGNTO) * RTA_ALIGNTO
+                i += align_to(attr_len, RTA_ALIGNTO)
                 logger.debug('rt_attr {} {}'.format(attr_len, attr_type))
 
             if addr is None:
                 logger.debug('not address in RTM message')
-                offset += ((msg_len + 1) // NLM_HDR_ALIGNTO) * NLM_HDR_ALIGNTO
+                offset += align_to(msg_len, NLM_HDR_ALIGNTO)
                 continue
 
-            iface = self.interfaces[ifa_idx]
-            if h_type == self.RTM_NEWADDR:
-                self.handle_new_address(addr, ifa_family, iface)
-            elif h_type == self.RTM_DELADDR:
-                self.handle_deleted_address(addr, ifa_family, iface)
+            if ifa_idx in self.interfaces:
+                iface = self.interfaces[ifa_idx]
+                if h_type == self.RTM_NEWADDR:
+                    self.handle_new_address(addr, ifa_family, iface)
+                elif h_type == self.RTM_DELADDR:
+                    self.handle_deleted_address(addr, ifa_family, iface)
+            else:
+                logger.debug('unknown interface index: {}'.format(ifa_idx))
 
-            offset += ((msg_len + 1) // NLM_HDR_ALIGNTO) * NLM_HDR_ALIGNTO
+            offset += align_to(msg_len, NLM_HDR_ALIGNTO)
 
     def cleanup(self):
         self.selector.unregister(self.socket)
@@ -1332,7 +1414,7 @@ class RouteSocketAddressMonitor(NetworkAddressMonitor):
                     if_name = (buf[off_name:off_name + name_len]).decode()
                     intf = self.add_interface(if_name, idx, idx)
 
-            offset += (((sa_len + SA_ALIGNTO - 1) // SA_ALIGNTO) * SA_ALIGNTO
+            offset += (align_to(sa_len, SA_ALIGNTO)
                        if sa_len > 0 else SA_ALIGNTO)
             addr_type_idx = addr_type_idx << 1
 
@@ -1442,9 +1524,17 @@ def parse_args():
         '-o', '--no-host',
         help='disable server mode operation (host will be undiscoverable)',
         action='store_true')
+    parser.add_argument(
+        '-V', '--version',
+        help='show version number and exit',
+        action='store_true')
 
     args = parser.parse_args(sys.argv[1:])
 
+    if args.version:
+        print('wsdd - Web Service Discovery Daemon, v{}'.format(WSDD_VERSION))
+        sys.exit(0)
+
     if args.verbose == 1:
         log_level = logging.INFO
     elif args.verbose > 1:
@@ -1496,7 +1586,7 @@ def send_outstanding_messages(block=False):
         addr = send_queue[-1][2]
         msg = send_queue[-1][3]
         try:
-            mch.send_socket.sendto(msg, addr)
+            mch.send(msg, addr)
         except Exception as e:
             logger.error('error while sending packet on {}: {}'.format(
                 mch.interface.name, e))
